//! Wrapper around the Solr JSON API

use std::collections::HashMap;
use std::io;

use reqwest::{Client, Url};
use serde::Deserialize;
use thiserror::Error;
use tracing::debug;

const DEFAULT_PORT: u16 = 8983;
const DEFAULT_PROTOCOL: &str = "http";

/// Errors
#[derive(Debug, Error)]
pub enum SolrError {
    /// Error in connecting to Solr
    #[error("error contacting to Solr: {0}")]
    Network(#[from] io::Error),

    /// Error in requests sent to Solr
    #[error("error in request to Solr: {0}")]
    RequestFailed(#[from] reqwest::Error),

    /// Some element of the url was not valid. This can happen for 2 reasons:
    ///     1. Invalid Solr instance information (e.g. wrong protocol)
    ///     2. Solr collection name creates an invalid url
    #[error("invalid url: {0}")]
    ParseUrlError(#[from] url::ParseError),

    /// Solr API has returned a non-successful status code
    #[error("Solr returned non-ok status: {status}\n{body}")]
    ResponseNotOk {
        /// status that was received
        status: reqwest::StatusCode,
        // TODO: this can likely be better described than just taking a string of the json response
        /// body of the response with information about the failure
        body: String,
    },
}

/// Wrapper around the Solr JSON API
///
/// Relevant docs:
/// * <https://solr.apache.org/guide/7_1/json-request-api.html>
/// * <https://solr.apache.org/guide/7_1/uploading-data-with-index-handlers.html>
#[derive(Debug, Clone)]
pub struct Solr {
    base_url: Url,
    client: Client,
}

impl Solr {
    /// Constructs a new `Solr` pointing to the indicated solr instance.
    /// If `port` is `None`, Solr's default port is used.
    pub fn new(
        host: String,
        port: Option<u16>,
        protocol: Option<String>,
    ) -> Result<Self, SolrError> {
        let port = port.unwrap_or(DEFAULT_PORT);
        let protocol = protocol.unwrap_or_else(|| DEFAULT_PROTOCOL.to_owned());

        let url = Url::parse(&format!("{}://{}:{}", protocol, host, port))?;
        Ok(Self {
            base_url: url,
            client: Client::new(),
        })
    }

    fn url(&self, collection: &str, method: &str) -> Result<Url, url::ParseError> {
        self.base_url
            .join(&format!("solr/{}/{}", collection, method))
    }

    /// Queries Solr `collection` for documents.
    /// Allows filtering via `facets`.
    pub async fn search(
        &self,
        collection: &str,
        start: usize,
        facets: &HashMap<&str, &str>,
    ) -> Result<SearchResponse, SolrError> {
        let url = self.url(collection, "select")?;

        // initialized here for lifetime purposes
        let start_str = format!("{}", start);
        let parameters = {
            let mut p = HashMap::from([("wt", "json"), ("q", "*:*"), ("start", &start_str)]);
            p.extend(facets);
            p
        };

        let req = self.client.get(url).query(&parameters);
        debug!("sending request:\n{:#?}", req);

        let res = req.send().await?;

        debug!("received response:\n{:#?}", res);
        if res.status() != reqwest::StatusCode::OK {
            Err(SolrError::ResponseNotOk {
                status: res.status(),
                body: res.json().await?,
            })
        } else {
            let response_body: SearchResponse = res.json().await?;
            Ok(response_body)
        }
    }

    /// Uploads `documents` into `collection` in Solr.
    pub async fn update(
        &self,
        collection: &str,
        documents: &[HashMap<&str, String>],
    ) -> Result<(), SolrError> {
        let url = self.url(collection, "update/json/docs")?;
        debug!("{}", url);
        let params = HashMap::from([
            ("wt", "json"), // sending json body
            ("overwrite", "true"),
            ("commit", "true"), // commit the data immediately
        ]);
        let req = self
            .client
            .post(url)
            .header(reqwest::header::CONTENT_TYPE, "application/json")
            .query(&params)
            .json(&documents);

        debug!("sending request:\n{:#?}", req);

        let res = req.send().await?;
        debug!("received response:\n{:#?}", res);

        if res.status() != reqwest::StatusCode::OK {
            Err(SolrError::ResponseNotOk {
                status: res.status(),
                body: res.text().await?,
            })
        } else {
            Ok(())
        }
    }
}

/// Response body from a `/search` request
#[derive(Debug, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct SearchResponse {
    /// Header
    pub response_header: ResponseHeader,
    /// Data
    pub response: ResponseBody,
}

/// The document data of the response
#[derive(Debug, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct ResponseBody {
    /// number of documents found that match this query
    pub num_found: usize,
    /// the starting index of this page of results
    pub start: i32,
    /// the documents in this page of results
    pub docs: Vec<FileDocument>,
}

/// Header of a solr response
#[derive(Debug, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct ResponseHeader {
    /// HTTP status code
    pub status: i32,
    /// Query parameters of the sent request
    pub params: HashMap<String, String>,
}

/// File document stored in Solr.
/// TODO: maybe this should be pulled out of here and defined by the caller?
#[derive(Debug, Deserialize)]
pub struct FileDocument {
    /// File's path
    pub file: String,
    /// When this document was inserted into Solr
    pub creation_time: String,
}
